#include <QFileDialog>
#include <QDir>
#include "addmacro.h"   // header file for addmacro dialog box
#include "addkey.h" // header file for add key dialog box
#include "addmac.h" // header file for add macro command dialog box
#include "targetwindow.h" // header file for target window dialog box

addmacro::addmacro(QWidget *parent)   // constructor for my main window
{
  setupUi(this); // this sets up GUI (all the stuff generated by Gt Designer and UIC)
  connect(pushButton_add,SIGNAL(clicked()),this,SLOT(AddCommand()));
  connect(pushButton_delete,SIGNAL(clicked()),this,SLOT(DeleteCommand()));
  connect(pushButton_up,SIGNAL(clicked()),this,SLOT(UpCommand()));
  connect(pushButton_down,SIGNAL(clicked()),this,SLOT(DownCommand()));
  connect(pushButton_browse,SIGNAL(clicked()),this,SLOT(BrowseWindow()));
}

void addmacro::AddCommand(void)
{
  switch (comboBox_cmdtype->currentIndex()) {
  case 0: // key
    AddKey();
    break;
  case 1: // mouse
    AddMouse();
    break;
  case 2: // launch
    AddLaunch();
    break;
  case 3: // macro
    AddMacro();
    break;
  }
}

int addmacro::SelectedCommand(void)
{
  if (listWidget_commands->selectedItems().size()!=1) {
    return -1;
  }
  return listWidget_commands->row(listWidget_commands->selectedItems()[0]);
}

void addmacro::DeleteCommand(void)
{
  int row;

  row=SelectedCommand();
  if (row>=0) {
    delete listWidget_commands->takeItem(row);
  }
}

void addmacro::UpCommand(void)
{
  int row;

  row=SelectedCommand();
  if (row>0) {
    listWidget_commands->insertItem(row-1,listWidget_commands->takeItem(row));
    listWidget_commands->setCurrentRow(row-1);
  }
}

void addmacro::DownCommand(void)
{
  int row;

  row=SelectedCommand();
  if (row>=0 && row<listWidget_commands->count()-1) {
    listWidget_commands->insertItem(row+1,listWidget_commands->takeItem(row));
    listWidget_commands->setCurrentRow(row+1);
  }
}

void addmacro::AddKey(void)
{
  addkey dialog(this);
  QString command;

  if (dialog.exec()) {
    command="KEY "+dialog.comboBox_key->currentText()+" ";
    if (dialog.checkBox_alt->isChecked()) {
      command+="+alt";
    }
    if (dialog.checkBox_ctrl->isChecked()) {
      command+="+ctrl";
    }
    if (dialog.checkBox_shift->isChecked()) {
      command+="+shift";
    }
    listWidget_commands->addItem(command);
  }
}

void addmacro::AddMouse(void)
{

}

void addmacro::AddLaunch(void)
{
  QString filename;
  static QString lastdir=QDir::homePath(); // last visited directory

  filename=QFileDialog::getOpenFileName(this,"File to Launch",lastdir);
  if (filename!="") {
    listWidget_commands->addItem(QString("LAUNCH ")+filename);
    QDir path(filename);
    lastdir=path.absolutePath(); // remember last visited directory
  }
}

void addmacro::AddMacro(void)
{
  addmac dialog(this);

  dialog.comboBox_macros->addItems(macros);
  if (dialog.exec()) {
    listWidget_commands->addItem(QString("MACRO ")+dialog.comboBox_macros->currentText());
  }
}

void addmacro::BrowseWindow(void)
{
  targetwindow dialog(this);

  if (dialog.exec()) {
    lineEdit_window->setText(dialog.treeWidget_windows->currentItem()->text(dialog.comboBox_use->currentIndex()));
  }
}
