#include <QtGui>    // this include ALL Qt class
#include "rfremote.h" // header file for main application windowsetval
#include "settings.h" // header file for settings dialog box
#include "addmacro.h" // header file for addmacro dialog box
#ifdef _WIN32
  #include "windows.h"    // header file for windows API
#else
#endif
/*
Note: After creating PRO file by qmake -project, the following lines must be added to the rfremote.pro file!!!
QT += xml
LIBS += -L. -lqextserialport
DEFINES += _TTY_POSIX_          #(linux only)
RC_FILE = rfremote.rc           #(windows only)

*/
/*
To create a proper Windows Icon file for the Windows EXE file:
First, create an ICO format bitmap file that contains the icon image.
This can be done with e.g. Microsoft Visual C++:
Select File|New, then select the File tab in the dialog that appears, and choose Icon.
(Note that you do not need to load your application into Visual C++; here we are only using the icon editor.)
Or use GIMP to create the ICO file!
Store the ICO file in your application's source code directory,
for example, with the name myappico.ico.
Then, create a text file called, say, myapp.rc in which you put a single line of text:

    IDI_ICON1               ICON    DISCARDABLE     "myappico.ico"

Finally, assuming you are using qmake to generate your makefiles, add this line to your myapp.pro file:

    RC_FILE = myapp.rc

Regenerate your makefile and your application. The .exe file will now be represented with your icon in Explorer.
*/

rfremote::rfremote(QWidget *parent)   // constructor for my main window
{
  QStringList cla;  // command line arguments

  setupUi(this); // this sets up GUI (all the stuff generated by Qt Designer and UIC)
  // set up all the signal/slot connections
  // 1st parameter is object name I gave to the button, etc.
  // SIGNAL specifies the signal emitted
  // SLOT specifies slot to connect to
  treeWidget_macros->header()->hide(); // hiding header from Qt Designer didn't work in 4.3.4
  connect(actionNew,SIGNAL(activated()),this,SLOT(NewMacros()));
  connect(actionExport_Macros,SIGNAL(activated()),this,SLOT(ExportMacros()));
  connect(actionImport_Macros,SIGNAL(activated()),this,SLOT(ImportMacros()));
  connect(actionSave,SIGNAL(activated()),this,SLOT(SaveChanges()));
  connect(action_About,SIGNAL(activated()),this,SLOT(doabout()));
  connect(action_Help,SIGNAL(activated()),this,SLOT(help()));
  CreateActions();
  connect(treeWidget_macros,SIGNAL(customContextMenuRequested(const QPoint &)),this,SLOT(TreeMenu(const QPoint &)));
  connect(actionSettings,SIGNAL(activated()),this,SLOT(Settings()));
  connect(pushButton_startstop,SIGNAL(clicked()),this,SLOT(Start()));
  cla=QCoreApplication::arguments();    // get list of command line arguments (first one is application exe file)
  if (cla.size()>=1) {
    HomePath=QFileInfo(cla[0]).absolutePath()+"/";
  }
  setWindowIcon(QIcon(HomePath+"RfRemote.png")); // make sure I can find the icon file
  IconLearned=QIcon(HomePath+"macro.png");
  IconUnlearned=QIcon(HomePath+"macroq.png");
  IconGroup=QIcon(HomePath+"group.png");
  readConfig(); // read rfremote.xml to get all persistant data
  QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","rfremote");
  readMacros(&settings); // read macros from default ini file
  DisplayTree(-1);
  changed=FALSE;
#ifndef _WIN32
// linux xwindows needs display and root window
  dpy=XOpenDisplay(NULL);
  root=RootWindow(dpy,DefaultScreen(dpy));
#endif
  learning=false;
  timer_popshow=0;
  label_popshow=NULL;
  autostart_flag=autostart; // schedule Start in timer if necessary (autominimize does not work if I call Start here
  // a signal/slot with a queued connection would be another solution to autostart/autominimize problem
  QTimer *timer=new QTimer(this);
  connect(timer, SIGNAL(timeout()), this, SLOT(Timer()));
  timer->start(50); // start executing Timer() every 50 milliseconds
}

void rfremote::CreateActions(void)
{
  addGlobalMacro=new QAction("Add Global Macro", this);
  connect(addGlobalMacro,SIGNAL(triggered()),this,SLOT(AddGlobalMacro()));
  addGroup=new QAction("Add Group", this);
  connect(addGroup,SIGNAL(triggered()),this,SLOT(AddGroup()));
  addMemberMacro=new QAction("Add Member Macro", this);
  connect(addMemberMacro,SIGNAL(triggered()),this,SLOT(AddMemberMacro()));
  editMacro=new QAction("Edit Macro", this);
  connect(editMacro,SIGNAL(triggered()),this,SLOT(EditMacro()));
  deleteMacro=new QAction("Delete Macro", this);
  connect(deleteMacro,SIGNAL(triggered()),this,SLOT(DeleteMacro()));
  learnMacro=new QAction("Learn Macro", this);
  connect(learnMacro,SIGNAL(triggered()),this,SLOT(LearnMacro()));
  executeMacro=new QAction("Execute Macro", this);
  connect(executeMacro,SIGNAL(triggered()),this,SLOT(ExecuteMacro()));
}

void rfremote::doabout(void)
{
  QMessageBox::about(this,"About RfRemote","<h1>RfRemote 2.00</h1><h2>by Keith Rubow</h2><a href=\"http://www.krubow.com/\">www.krubow.com</a>");
}

void rfremote::help(void)
{
  QDesktopServices::openUrl(QUrl(HomePath+"rfremote.html"));
}

// return the macro number whose learn data matches the last part of data
// only include member macros in active group
int rfremote::MatchMacro(QString data)
{
  int num,len,i;
  bool activegroup;
  struct macro *thismacro;

  activegroup=0;
  num=macrolist.size();
  for (i=0; i<num; i++) {
    thismacro=&macrolist[i];
    if (thismacro->type==MACRO_GROUP) {
      activegroup=(ActiveGroup==thismacro->name);
    }
    if (thismacro->type!=MACRO_MEMBER || activegroup) {
      len=macrolist[i].learn.size();
      if (len>0 && macrolist[i].learn==data.right(len)) { // macro with valid learn data that matches end of data
        return i; // return macro number
      }
    }
  }
  return -1; // no macro found
}

// timer executes every 50 ms
void rfremote::Timer(void)
{
  static QString data; // accumulated input data as hexadecimal string
  char ch;
  int n;
  static int quiettimer;

  if (autostart_flag) {
    autostart_flag=0;
    Start(); // autostart Caution: autominimize does not work if I call Start from rfremote constructor
  }
  if (timer_popshow) { // if popup window is active
    if (--timer_popshow==0) { // if display timer expired
      delete label_popshow; // hide and destroy the popup window
      label_popshow=NULL; // indicate no popup window exists
    }
  }
  if (CommPort.isOpen()) {
    quiettimer++;
    if (quiettimer>=4) {
      quiettimer=4;
    }
    while (CommPort.bytesAvailable()>0) {
      CommPort.readData(&ch,1); // get one character
      if (learning) {
        if (quiettimer>=4) {
          learndialog->lineEdit_learn->clear();
        }
        quiettimer=0;
        learndialog->lineEdit_learn->setText(learndialog->lineEdit_learn->text()+QString("%1").arg((unsigned char)ch,2,16,QChar('0'))); // append 2 digit hex string to end of learn data
      } else {
        quiettimer=0;
        data+=QString("%1").arg((unsigned char)ch,2,16,QChar('0')); // append 2 digit hex string to end of data
        data=data.right(256); // limit accumulated data to 128 bytes (256 hexadecimal characters)
//QMessageBox::information(this,"data is",data);
        n=MatchMacro(data); // look for a matching macro
        if (n>=0) {
          data.clear(); // clear accumulated data
          ExecuteIt(&macrolist[n]); // execute the macro that matched
        }
      }
    }
  } else {
    quiettimer=0;
  }
}

void rfremote::Start(void)
{
  if (CommPort.isOpen()) {
    CommPort.close();
    pushButton_startstop->setText("Start");
    ActiveGroup=""; // no active group
  } else {
    OpenPort();
    if (CommPort.isOpen()) {
      pushButton_startstop->setText("Stop");
      if (!learning) {
        if (autominimize) {
          setWindowState(windowState()|Qt::WindowMinimized); // minimize window
        }
        ExecuteMac("Startup");
      }
    }
  }
}

void rfremote::OpenPort(void)
{
  BaudRateType baudrate[]={BAUD300,BAUD600,BAUD1200,BAUD2400,BAUD4800,BAUD9600,BAUD19200,BAUD38400,BAUD57600,BAUD115200};

  CommPort.setPortName(port);
  CommPort.setDataBits(DATA_8);
  CommPort.setParity(PAR_NONE);
  CommPort.setFlowControl(FLOW_OFF);
  CommPort.setStopBits(STOP_1);
  CommPort.setBaudRate(baudrate[baud]);
  CommPort.open();
  if (!CommPort.isOpen()) {
    QMessageBox::information(this,"Error",QString("Cannot open %1").arg(port));
  }
}

// return number of selected macro, or -1 if none selected
int rfremote::SelectedNumber(void)
{
  int i;

  if (treeWidget_macros->selectedItems().size()==1) {
    i=treeWidget_macros->selectedItems()[0]->data(0,32).toInt();
    return i;
  }
  return -1;
}

// return type of selected macto, or -1 if none selected
int rfremote::SelectedType(void)
{
  int i;

  if ((i=SelectedNumber())>=0) {
    i=macrolist[i].type;
  }
  return i;
}

void rfremote::TreeMenu(const QPoint &pos)
{
//  QMessageBox::information(this,"",QString("Popup at %1,%2").arg(pos.x()).arg(pos.y()));
  QMenu menu(this);
  int type;

  type=SelectedType();
  menu.addAction(addGlobalMacro);
  menu.addAction(addGroup);
  menu.addAction(addMemberMacro);
  menu.addSeparator();
  menu.addAction(editMacro);
  menu.addAction(deleteMacro);
  menu.addSeparator();
  menu.addAction(learnMacro);
  menu.addAction(executeMacro);
  
  menu.actions()[2]->setEnabled(type==MACRO_GROUP || type==MACRO_MEMBER); // restrict add member macro
  menu.actions()[4]->setEnabled(type>=0); // restrict edit macro
  menu.actions()[5]->setEnabled(type>=0); // restrict delete macro
  menu.actions()[7]->setEnabled(type>=0); // restrict learn macro
  menu.actions()[8]->setEnabled(type>=0); // restrict execute macro
  menu.exec(treeWidget_macros->mapToGlobal(pos));
}

// given pointer to new macro
// (and entry position of parent if it a member macro)
// insert new macro at proper point in macrolist to keep the list sorted
// return -1 if error
// return index of new entry if entered successfully
// check for duplicate macro names as follows:
// NO macro may duplicate the name of a global or group macro
// a member macro may not duplicate the name of another member macro in the same group
int rfremote::AddMacro(struct macro *newmacro, int parent)
{
  int i,n;

  n=macrolist.size();
  for (i=0; i<n; i++) { // check for duplicate global or group
    if (macrolist[i].type==MACRO_GLOBAL || macrolist[i].type==MACRO_GROUP) {
      if (macrolist[i].name==newmacro->name) {
        return -1; // error, name is duplicate of existing global or group macro
      }
    }
  }
  switch (newmacro->type) {
  case MACRO_GLOBAL: // global macros always come first
    for (i=0; i<n; i++) { // find the insert point (end of globals or higher global name)
      if (macrolist[i].type!=MACRO_GLOBAL || newmacro->name<macrolist[i].name) {
        break;
      }
    }
    break;
  case MACRO_GROUP:
    for (i=0; i<n; i++) { // find insert point (higher group name or end of list)
      if (macrolist[i].type==MACRO_GROUP && newmacro->name<macrolist[i].name) {
        break;
      }
    }
    break;
  case MACRO_MEMBER:
    if (parent<0 || parent>=n || macrolist[parent].type!=MACRO_GROUP) {
      return -1; // invalid parent
    }
    for (i=parent+1; i<n && macrolist[i].type==MACRO_MEMBER; i++) { // find insert point
      if (newmacro->name==macrolist[i].name) {
        return -1; // duplicate member macro name
      }
      if (newmacro->name<=macrolist[i].name) {
        break;
      }
    }
    break;
  }
  macrolist.insert(i,*newmacro);
  changed=TRUE;
  return i;
}

// delete macro number n in macrolist (and all member macros if it is a group macro)
void rfremote::DelMacro(int n)
{
  int type;

  type=macrolist[n].type;
  macrolist.removeAt(n);
  changed=TRUE;
  if (type==MACRO_GROUP) {
    while (macrolist.size()>n && macrolist[n].type==MACRO_MEMBER) {
      macrolist.removeAt(n);
    }
  }
}

// display tree and make macro item n visible
void rfremote::DisplayTree(int n)
{
  int topcount,childcount,i,len;
  QTreeWidgetItem *newitem;

  treeWidget_macros->clear(); // empty the tree widget
  len=macrolist.size();
  topcount=0;
  childcount=0;
  for (i=0; i<len; i++) {
    newitem=new QTreeWidgetItem;
    newitem->setData(0,32,QVariant(i)); // data refers back to macrolist entry
    switch (macrolist[i].type) {
    case MACRO_GLOBAL:
    case MACRO_GROUP:
      newitem->setText(0,macrolist[i].name);
      if (macrolist[i].type==MACRO_GROUP) {
        newitem->setIcon(0,IconGroup);
      } else {
        if (macrolist[i].learn=="") {
          newitem->setIcon(0,IconUnlearned);
        } else {
          newitem->setIcon(0,IconLearned);
        }
      }
      treeWidget_macros->insertTopLevelItem(topcount++,newitem);
      childcount=0;
      if (i==n) {
        treeWidget_macros->scrollToItem(treeWidget_macros->topLevelItem(topcount-1)); // make item visible
      }
      break;
    case MACRO_MEMBER:
      newitem->setText(0,macrolist[i].name);
      if (macrolist[i].learn=="") {
        newitem->setIcon(0,IconUnlearned);
      } else {
        newitem->setIcon(0,IconLearned);
      }
      treeWidget_macros->topLevelItem(topcount-1)->insertChild(childcount++,newitem);
      if (i==n) {
        treeWidget_macros->scrollToItem(treeWidget_macros->topLevelItem(topcount-1)->child(childcount-1));
      }
      break;
    }
  }
}

// return a list of macros available to macro number n
// this is all global and group macros, and members of group if this is a group or member macro
QStringList rfremote::AvailableMacros(int n)
{
  QStringList macros;
  int parent,group,i,len;

  len=macrolist.size();
  if (n<len) {
    if (macrolist[n].type!=MACRO_GLOBAL) {
      for (parent=n; parent>=0; parent--) { // find parent
        if (macrolist[parent].type==MACRO_GROUP) {
          break;
        }
      }
    } else {
      parent=-1;
    }
    group=-1;
    for (i=0; i<len; i++) {
      switch (macrolist[i].type) {
      case MACRO_GROUP:
        group=i;
      case MACRO_GLOBAL:
        macros<<macrolist[i].name;
        break;
      case MACRO_MEMBER:
        if (group==parent) {
          macros<<macrolist[i].name;
        }
      }
    }
  }
  return macros;
}

void rfremote::AddGlobalMacro(void)
{
  addmacro dialog(this);
  struct macro newmacro;
  int i,n,len;
 
  newmacro.type=MACRO_GLOBAL;
  newmacro.commands=QStringList();
  newmacro.loopcnt=0;
  dialog.macros=AvailableMacros(0);
  for (;;) {
    dialog.lineEdit_name->setFocus();
    if (dialog.exec()) {
      newmacro.name=dialog.lineEdit_name->text();
      newmacro.loop=dialog.checkBox_loop->isChecked();
      newmacro.window=dialog.lineEdit_window->text();
      newmacro.commands.clear();
      len=dialog.listWidget_commands->count();
      for (i=0; i<len; i++) {
        newmacro.commands<<dialog.listWidget_commands->item(i)->text();
      }
      if ((n=AddMacro(&newmacro,0))<0) {
        QMessageBox::warning(this,"","Error, duplicate macro name!");
      } else {
        DisplayTree(n);
        return;
      }
    } else {
      return;
    }
  }
}

void rfremote::AddGroup(void)
{
  addmacro dialog(this);
  struct macro newmacro;
  int n;
 
  newmacro.type=MACRO_GROUP;
  newmacro.commands=QStringList();
  newmacro.loop=0;
  newmacro.loopcnt=0;
  dialog.groupBox_commands->setEnabled(FALSE);
  dialog.checkBox_loop->setEnabled(FALSE);
  dialog.lineEdit_window->setEnabled(FALSE);
  dialog.pushButton_browse->setEnabled(FALSE);

  dialog.macros=AvailableMacros(0);
  for (;;) {
    dialog.lineEdit_name->setFocus();
    if (dialog.exec()) {
       newmacro.name=dialog.lineEdit_name->text();
       if ((n=AddMacro(&newmacro,0))<0) {
         QMessageBox::warning(this,"","Error, duplicate macro name!");
       } else {
         DisplayTree(n);
         return;
       }
    } else {
      return;
    }
  }
}

void rfremote::AddMemberMacro(void)
{
  int parentnum,i,n,len;
  addmacro dialog(this);
  struct macro newmacro;

  for (parentnum=SelectedNumber(); parentnum>=0; parentnum--) {
    if (macrolist[parentnum].type==MACRO_GROUP) {
      break;
    }
  }
  if (parentnum<0) {
    return;
  }
  newmacro.type=MACRO_MEMBER;
  newmacro.commands=QStringList();
  newmacro.loopcnt=0;
  dialog.macros=AvailableMacros(parentnum);
  for (;;) {
    dialog.lineEdit_name->setFocus();
    if (dialog.exec()) {
      newmacro.name=dialog.lineEdit_name->text();
      newmacro.loop=dialog.checkBox_loop->isChecked();
      newmacro.window=dialog.lineEdit_window->text();
      newmacro.commands.clear();
      len=dialog.listWidget_commands->count();
      for (i=0; i<len; i++) {
        newmacro.commands<<dialog.listWidget_commands->item(i)->text();
      }
      if ((n=AddMacro(&newmacro,parentnum))<0) {
        QMessageBox::warning(this,"","Error, duplicate macro name!");
      } else {
        DisplayTree(n);
        return;
      }
    } else {
      return;
    }
  }
}

// caution, editing the name of a macro can be difficult
// if it is a group, entire group must be moved
// must not create duplicate names
void rfremote::EditMacro(void)
{
  addmacro dialog(this);
  struct macro newmacro;
  int i,n,parent,len,temp,newpos;
  QList<struct macro> savemacrolist;

  n=SelectedNumber();
  newmacro=macrolist[n];
  dialog.lineEdit_name->setText(newmacro.name);
  if (macrolist[n].type==MACRO_GROUP) {
    dialog.groupBox_commands->setEnabled(FALSE);
    dialog.checkBox_loop->setEnabled(FALSE);
    dialog.lineEdit_window->setEnabled(FALSE);
    dialog.pushButton_browse->setEnabled(FALSE);
  } else {
    dialog.checkBox_loop->setChecked(newmacro.loop);
    dialog.lineEdit_window->setText(newmacro.window);
    dialog.listWidget_commands->addItems(newmacro.commands);
  }
  dialog.macros=AvailableMacros(n);
  for (;;) {
    if (dialog.exec()) {
      newmacro.name=dialog.lineEdit_name->text();
      if (newmacro.type!=MACRO_GROUP) {
        newmacro.loop=dialog.checkBox_loop->isChecked();
      }
      newmacro.window=dialog.lineEdit_window->text();
      newmacro.commands.clear();
      len=dialog.listWidget_commands->count();
      for (i=0; i<len; i++) {
        newmacro.commands<<dialog.listWidget_commands->item(i)->text();
      }
      savemacrolist.clear();
      savemacrolist=macrolist; // save a copy of the macro list
      macrolist.clear(); // clear the macro list
      len=savemacrolist.size();
      parent=0;
      newpos=-1;
      for (i=0; i<len; i++) { // rebuild the macro list
        if (i==n) {
          temp=AddMacro(&newmacro,parent);
          newpos=temp; // this is the position of the edited macro
        } else {
          temp=AddMacro(&(savemacrolist[i]),parent);
          if (temp<=newpos) {
            newpos++; // adjust position of edited macro if another macro got inserted before it
          }
        }
        if (temp<0) {
          QMessageBox::warning(this,"","Error, duplicate macro name!");
          macrolist.clear();
          macrolist=savemacrolist;
          savemacrolist.clear();
          break;
        }
        if (macrolist[i].type==MACRO_GROUP) {
          parent=i;
        }
      }
      if (i==len) { // no duplicate name, list created OK
        savemacrolist.clear();
        DisplayTree(newpos);
        break;
      }
    } else {
      break;
    }
  }
}

void rfremote::DeleteMacro(void)
{
  int n,parentnum;

  n=SelectedNumber();
  parentnum=-1;
  if (macrolist[n].type==MACRO_MEMBER) {
    for (parentnum=n; parentnum>=0; parentnum--) {
      if (macrolist[parentnum].type==MACRO_GROUP) {
        break;
      }
    }
  }
  DelMacro(SelectedNumber());
  DisplayTree(parentnum);
  if (parentnum>=0) { // find and expand parent if necessary
    QTreeWidgetItemIterator item(treeWidget_macros);
    while (*item) { // iterate over all items in the tree to find the parent of deleted member
      if ((*item)->data(0,32).toInt()==parentnum) {
        treeWidget_macros->expandItem(*item); // expand the parent
        return;
      }
      ++item;
    }
  }
}

void rfremote::LearnMacro(void)
{
  learn dialog(this);
  bool wasopen;
  int i,n,parentn,len;
  bool mygroupflag;
  struct macro *thismacro;

  learndialog=&dialog;
  learning=true;
  mygroupflag=false;
  wasopen=CommPort.isOpen(); // remember if comm port was open or not
  if (!wasopen) {
    OpenPort(); // open port if necessary
  }
  if (CommPort.isOpen()) { // if port opened OK
    n=SelectedNumber();
    thismacro=&macrolist[n];
    parentn=-1;
    len=macrolist.size();
    if (thismacro->type==MACRO_MEMBER) { // find parent of this member macro
      for (i=n-1; i>=0; i--) { // scan backwards to previous group
        if (macrolist[i].type==MACRO_GROUP) {// found it
          break;
        }
      }
      parentn=i;
    }
    dialog.setWindowTitle(QString("Learning %1").arg(thismacro->name));
    dialog.lineEdit_learn->setText(thismacro->learn); // start by displaying current learn text
    for (;;) { // repeat until valid learn data, or cancel
      if (dialog.exec()) {
        // check for duplicate learn string
        // global macro may not duplicate any macro
        // member macro may not duplicate any global macro, or any member on same group
        i=len;
        if (dialog.lineEdit_learn->text().size()>0) {
          for (i=0; i<len; i++) {
            if (i!=n) { // no need to check the nacro being learned
              if (macrolist[i].type==MACRO_GROUP) {
                mygroupflag=(i==parentn); // flag that I am in same group as macro being learned
              } else {
                if (thismacro->type==MACRO_GLOBAL || (macrolist[i].type==MACRO_GLOBAL || mygroupflag)) {
                  if (dialog.lineEdit_learn->text()==macrolist[i].learn) {
                    QMessageBox::information(this,"Error","Duplicate learn, button already in use!");
                    break;
                  }
                }
              }
            }
          }
        }
        if (i>=len) {
          thismacro->learn=dialog.lineEdit_learn->text(); // should check for duplicate learn text
          changed=true;
          break;
        }
      } else {
        break; // cancel
      }
    }
    if (!wasopen) {
      CommPort.close(); // close port if it was not open
    }
    DisplayTree(-1);
  }
  learning=false;
}

// show a popup window with macro name in selected position/size/color
void rfremote::Show(QString name)
{
//  QMessageBox::information(this,"debug",QString("Show macro %1 not implemented yet").arg(name));
  QRect screenrect; // size of screen
  int top,bottom,left,right;
  QSize popsize; // size of popup window
  int x,y,loc;

  if (showpos==0) {
    return; // do not display
  }
  loc=showpos-1;
// loc is position on screen, 0-8 means topleft, topcenter, topright, middleleft, etc. as follows
//0 1 2
//3 4 5
//6 7 8
  if (label_popshow!=NULL) {
    delete label_popshow; // reusing existing window did not seem to work reliably, so destroy it and create new one
  }
  label_popshow=new QLabel(name,0,Qt::SubWindow|Qt::FramelessWindowHint); // create label with text name as frameless subwindow
  label_popshow->setStyleSheet(QString("background-color:white; font-size:%1pt; color:#%2;").arg(showfontsize).arg(showfontcolor,6,16,QChar('0')));
  screenrect=QApplication::desktop()->screenGeometry(QApplication::desktop()->primaryScreen()); // rectangle of primary screen
  top=screenrect.top(); // topmost display
  bottom=screenrect.bottom(); // bottommost display
  left=screenrect.left(); // leftmost display
  right=screenrect.right(); // rightmost display
  popsize=label_popshow->sizeHint().boundedTo(QSize(right-left,bottom-top)); // sizehint is desired size of label, limit to screen size
  label_popshow->resize(popsize); // set label size to smaller of size hint and screen
  // now calculate limits of display position
  bottom-=popsize.height(); // bottommost position
  if (bottom<top) {
    bottom=top;
  }
  right-=popsize.width(); // rightmost position
  if (right<left) {
    right=left;
  }
  // now calculate actual display position
  x=left+(right-left)*(loc%3)/2; // display position depends on loc
  y=top+(bottom-top)*(loc/3)/2;
  label_popshow->move(x,y);
  label_popshow->show(); // show popup window at selected position
  timer_popshow=20*2; // set timer for 2 seconds
  printf("top=%d, bottom=%d, left=%d, right=%d, loc=%d, x=%d, y=%d\n",top,bottom,left,right,loc,x,y);
}

//wrapper to avoid always having to check for null pointer
void myxfree(void *ptr)
{
  if (ptr!=NULL) {
    XFree(ptr); // XFree cannot take a NULL pointer, which was really dumb!
  }
}

// find window by name (current window if name is null or blank)
// return 0 if not found
Window rfremote::FindWindow(Window top, const char *name)
{
  Window *children,foo;
  unsigned int nc;
  char *wname;
  XClassHint classhint;
  int revert_to_return;
  int len;
  int xpos,ypos;
  unsigned int xsize,ysize,border,depth;
  bool hastitle;

  if (name==NULL || *name==0) { // no name given
    XGetInputFocus(dpy, &foo, &revert_to_return);
    return(foo); // return current window
  }
  len=strlen(name); // length of name will always be >0
//  hastitle=false;
//  if (XFetchName(dpy,top,&wname)) {
//    if (wname && *wname) {
//      hastitle=true;
//    }
//    myxfree(wname);
//  }
  if (XGetGeometry(dpy,top,&foo,&xpos,&ypos,&xsize,&ysize,&border,&depth)==0) {
    xpos=ypos=0; xsize=ysize=border=0;
  }
  if (xsize*ysize>100) { // only check windows with reasonably large size
    hastitle=false;
    if (XFetchName(dpy,top,&wname)) { // get window name
      if (wname && *wname) {
        hastitle=true;
      }
      if (strncmp(wname,name,len)==0)  { // if name matches
        myxfree(wname);
//QMessageBox::information(this,"debug",QString("Found %1 by window name %2").arg(name).arg((int)top));
        return top;
      }
      myxfree(wname);
    }
    if (hastitle) { // only consider this window if it had a title
// I do not currently identify windows by IconName
//      if (XGetIconName(dpy,top,&wname)) { // get icon name
//        if (strncmp(wname,name,len)==0)  { // if icon name matches
//          myxfree(wname);
//QMessageBox::information(this,"debug",QString("Found %1 by icon name %2").arg(name).arg((int)top));
//          return top;
//        }
//        myxfree(wname);
//      }
      if (XGetClassHint(dpy,top,&classhint)) { // get class hint
        if(strcmp(classhint.res_class,name)==0)  { // if class hint resource name matches exactly
          myxfree(classhint.res_name);
          myxfree(classhint.res_class);
//QMessageBox::information(this,"debug",QString("Found %1 by class hint %2").arg(name).arg((int)top));
          return top; // found it
        }
        myxfree(classhint.res_name);
        myxfree(classhint.res_class);
      }
    }
  }
  if (XQueryTree(dpy,top,&foo,&foo,&children,&nc)==0 || children==NULL) {
    return 0; // not found
  }
  for (; nc>0; nc--)  { // check all children recursively
    top=FindWindow(children[nc-1],name);
    if (top) { // found it
      myxfree(children);
      return top;
    }
  }
  myxfree(children);
  return 0; // not found
}

Time fake_timestamp()
{
  QTime t;

  t=QTime::currentTime();
  return (Time)(((t.hour()*60+t.minute())*60+t.second())*1000+t.msec());
}

void rfremote::SendKey(const char *keyname, bool ctrl, bool shift, bool alt, Window w)
{
  XEvent xevent;

  xevent.xkey.display=dpy;
  xevent.xkey.root=root;
  xevent.xkey.subwindow=None;
  xevent.xkey.x=1;
  xevent.xkey.y=1;
  xevent.xkey.x_root=1;
  xevent.xkey.y_root=1;
  xevent.xkey.same_screen=True;
  xevent.xkey.state=0;
  if (ctrl) {
    xevent.xkey.state|=ControlMask;
  }
  if (shift) {
    xevent.xkey.state|=ShiftMask;
  }
  if (alt) {
    xevent.xkey.state|=Mod1Mask;
  }
  xevent.xkey.window=w;
  xevent.xkey.subwindow=0;
  xevent.xkey.keycode=XKeysymToKeycode(dpy,XStringToKeysym(keyname));
  xevent.xkey.type=KeyPress;
  xevent.xkey.time=fake_timestamp();
  XSendEvent(dpy,w,True,KeyPressMask,&xevent);
  xevent.xkey.type=KeyRelease;
  xevent.xkey.time+=2000; // fake timestamp 2sec later
  XSendEvent(dpy,w,True,KeyReleaseMask,&xevent);
  XSync(dpy,True);
}

void rfremote::ExecuteKey(QString key, QString window)
{
  bool alt,ctrl,shift;
  int pos;
  Window w;
  QString translated_key;
  char const *transkey[]={ // table of key names to translate into xevent key names if necessary (no translation if no entry)
    "`","grave",
    "-","minus",
    "=","equal",
    "[","bracketleft",
    "]","bracketright",
    ";","semicolon",
    "'","apostrophe",
    ",","comma",
    ".","period",
    "/","slash",
    "BACKSPACE","BackSpace",
    "DELETE","Delete",
    "DOWNARROW","Down",
    "END","End",
    "ENTER","Return",
    "ESC","Escape",
    "HOME","Home",
    "INSERT","Insert",
    "LEFTARROW","Left",
    "NUMPAD0","KP_0",
    "NUMPAD1","KP_1",
    "NUMPAD2","KP_2",
    "NUMPAD3","KP_3",
    "NUMPAD4","KP_4",
    "NUMPAD5","KP_5",
    "NUMPAD6","KP_6",
    "NUMPAD7","KP_7",
    "NUMPAD8","KP_8",
    "NUMPAD9","KP_9",
    "NUMPAD*","KP_Multiply",
    "NUMPAD/","KP_Divide",
    "NUMPAD+","KP_Add",
    "NUMPAD-","KP_Subtract",
    "NUMPAD.","KP_Decimal",
    "NUMPADENTER","KP_Enter",
    "PAGEUP","Page_Up",
    "PAGEDOWN","Page_Down",
    "PAUSE/BREAK","Pause",
    "PRINTSCREEN","Print", // is this correct?
    "RIGHTARROW","Right",
    "SPACE","space",
    "TAB","Tab",
    "UPARROW","Up",
    NULL,NULL};
  const char **ptr;    

  alt=key.indexOf("+alt")>0;
  ctrl=key.indexOf("+ctrl")>0;
  shift=key.indexOf("+shift")>0;
  pos=key.indexOf(" ");
  if (pos>0) {
    key=key.left(pos);
  }
  translated_key=key;
  for (ptr=transkey; *ptr; ptr++,ptr++) {
    if (key==*ptr) { // found key name
      translated_key=*++ptr; // translated key name
      break;
    }
  }
  w=FindWindow(root,window.toAscii());
  if (w) {
    XSetInputFocus(dpy,w,RevertToNone,CurrentTime); // Qt apps like Wplayer don't receive synthetic keys unless they have focus
    XFlush(dpy); // this seemed to be necessary to get the focus set before sending the synthetic key
    SendKey(translated_key.toAscii(),ctrl,shift,alt,w);
  }
}

// find macro name (global, group, or member of active group)
// return pointer to macro, or NULL if not found
struct rfremote::macro *rfremote::FindMacro(QString name)
{
  int i,n;
  struct macro *thismacro;
  bool activegroup;

  n=macrolist.size();
  activegroup=false;
  for (i=0; i<n; i++) {
    thismacro=&macrolist[i];
    if (thismacro->type==MACRO_GROUP) {
      activegroup=(ActiveGroup==thismacro->name);
    }
    if (thismacro->type!=MACRO_MEMBER || activegroup) {
      if (thismacro->name==name) {
        return thismacro;
      }
    }
  }
  return NULL;
}

void rfremote::ExecuteMac(QString macname)
{
  struct macro *macro;

  macro=FindMacro(macname); // find the macro
  if (macro) {
    ExecuteIt(macro); // execute it
  }
}

void rfremote::ExecuteLaunch(QString launchcmd)
{
  if (!QProcess::startDetached(launchcmd)) { // try to execute it as an application or script
    QDesktopServices::openUrl(QUrl(launchcmd)); // try to launch the file with associated application
  }
}

void rfremote::ExecuteCmd(QString cmd, QString window)
{
  char const *commands[]={"KEY","MACRO","LAUNCH",NULL};
  int i,len;

  i=0;
  while (commands[i]!=NULL) {
    len=strlen(commands[i]);
    if (cmd.left(len)==commands[i]) {
      switch (i) {
      case 0: // KEY
        ExecuteKey(cmd.remove(0,len+1),window);
        break;
      case 1: // MACRO
        ExecuteMac(cmd.remove(0,len+1));
        break;
      case 2: // LAUNCH
        ExecuteLaunch(cmd.remove(0,len+1));
        break;
      }
      break;
    }
    i++;
  }
}

void rfremote::ExecuteIt(struct macro *macro)
{
  int ncmds,i;

  Show(macro->name); // display the macro name in a popup window if enabled
  if (macro->type==MACRO_GROUP) {
    ActiveGroup=macro->name; // executing a group macro makes it the active group
    return;
  }
  ncmds=macro->commands.size();
  if (macro->loop) { // execute one command at a time
    if (macro->loopcnt>=ncmds) {
      macro->loopcnt=0;
    }
    ExecuteCmd(macro->commands[macro->loopcnt++],macro->window);
  } else {
    for (i=0; i<ncmds; i++) { // execute all commands in sequence
      ExecuteCmd(macro->commands[i],macro->window);
    }
  }
}

void rfremote::ExecuteMacro(void)
{
  ExecuteIt(&macrolist[SelectedNumber()]);
}

void rfremote::Settings(void)
{
  settings dialog(this);

  dialog.comboBox_port->lineEdit()->setText(port);
  dialog.comboBox_baud->setCurrentIndex(baud);
  dialog.comboBox_show->setCurrentIndex(showpos);
  dialog.spinBox_fontsize->setValue(showfontsize);
  dialog.setFontColor(showfontcolor);
  dialog.checkBox_autostart->setChecked(autostart);
  dialog.checkBox_autominimize->setChecked(autominimize);
  if (dialog.exec()) {
    port=dialog.comboBox_port->lineEdit()->text();
    baud=dialog.comboBox_baud->currentIndex();
    showpos=dialog.comboBox_show->currentIndex();
    showfontsize=dialog.spinBox_fontsize->value();
    showfontcolor=dialog.getFontColor();
    autostart=dialog.checkBox_autostart->isChecked();
    autominimize=dialog.checkBox_autominimize->isChecked();
    writeConfig();  // save changes
  }
}

void rfremote::readConfig(void)
{
  QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","rfremote"); // constructor, force INI file storage
  port=settings.value("port").toString();
  baud=settings.value("baud",3).toInt();
  showpos=settings.value("showpos",0).toInt();
  showfontsize=settings.value("showfontsize",48).toInt();
  showfontcolor=settings.value("showfontcolor",0).toInt();
  autostart=settings.value("autostart",0).toInt();
  autominimize=settings.value("autominimize",0).toInt();
}

void rfremote::writeConfig(void)
{
  QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","rfremote"); // constructor, force INI file storage
  settings.setValue("port",port);
  settings.setValue("baud",baud);
  settings.setValue("showpos",showpos);
  settings.setValue("showfontsize",showfontsize);
  settings.setValue("showfontcolor",showfontcolor);
  settings.setValue("autostart",autostart);
  settings.setValue("autominimize",autominimize);
}

// read macro definitions from the given QSettings object
void rfremote::readMacros(QSettings *settings)
{
  int macrosize,commandsize,macron,commandn;
  struct macro newmacro;

  newmacro.commands=QStringList();
  macrosize=settings->beginReadArray("macros");
  for (macron=0; macron<macrosize; macron++) {
    settings->setArrayIndex(macron);
    newmacro.name=settings->value("name").toString();
    newmacro.window=settings->value("window").toString();
    newmacro.learn=settings->value("learn").toString();
    newmacro.type=settings->value("type").toInt();
    newmacro.loop=settings->value("loop").toInt();
    newmacro.commands.clear();
    commandsize=settings->beginReadArray("commands");
    for (commandn=0; commandn<commandsize; commandn++) {
      settings->setArrayIndex(commandn);
      newmacro.commands.append(settings->value("command").toString());
    }
    settings->endArray(); // end of commands
    macrolist.append(newmacro);
  }
  settings->endArray(); // end of macros
}

// save all macros in given QSettings object
void rfremote::writeMacros(QSettings *settings)
{
  int macrosize,commandsize,macron,commandn;
  settings->beginGroup("macros");
  settings->remove(""); // remove all old macro keys from ini file
  settings->endGroup();
  settings->beginWriteArray("macros");
  macrosize=macrolist.size();
  for (macron=0; macron<macrosize; macron++) {
    settings->setArrayIndex(macron);
    settings->setValue("name",macrolist[macron].name);
    settings->setValue("window",macrolist[macron].window);
    settings->setValue("learn",macrolist[macron].learn);
    settings->setValue("type",macrolist[macron].type);
    settings->setValue("loop",macrolist[macron].loop);
    settings->beginWriteArray("commands");
    commandsize=macrolist[macron].commands.size();
    for (commandn=0; commandn<commandsize; commandn++) {
      settings->setArrayIndex(commandn);
      settings->setValue("command",macrolist[macron].commands[commandn]);
    }
    settings->endArray(); // end of commands
  }
  settings->endArray(); // end of macros
}

void rfremote::NewMacros(void)
{
  macrolist.clear(); // delete all macro definitions
  DisplayTree(-1);
  changed=FALSE;
}

void rfremote::ExportMacros(void)
{
  QString filename;
  static QString lastdir=QDir::homePath(); // last visited directory
  QFileInfo fileinfo;

  filename=QFileDialog::getSaveFileName(this,"Export Macros",lastdir,"INI files (*.ini)");
  if (filename!="") {
    fileinfo.setFile(filename);
    if (fileinfo.suffix().toLower()!="ini") {
      filename+=".ini"; // ensure ini file type
    }
    QSettings settings(filename,QSettings::IniFormat);
    writeMacros(&settings); // write macros to specified ini file
    QDir path(filename);
    lastdir=path.absolutePath(); // remember last visited directory
  }
}

void rfremote::ImportMacros(void)
{
  QString filename;
  static QString lastdir=QDir::homePath(); // last visited directory

  filename=QFileDialog::getOpenFileName(this,"Import Macros",lastdir,"INI files (*.ini)");
  if (filename!="") {
    macrolist.clear(); // delete all macro definitions
    QSettings settings(filename,QSettings::IniFormat);
    readMacros(&settings); // read macros from specified ini file
    DisplayTree(-1);
    changed=TRUE;
    QDir path(filename);
    lastdir=path.absolutePath(); // remember last visited directory
  }
}

void rfremote::SaveChanges(void)
{
  QSettings settings(QSettings::IniFormat,QSettings::UserScope,"krubow","rfremote");
  writeMacros(&settings); // save macros to default ini file
  changed=0;
}

void rfremote::closeEvent(QCloseEvent *event)
{
  if (changed) {
    if (QMessageBox::question(this,"Macros changed","Save changes to macros?",QMessageBox::Yes|QMessageBox::No,QMessageBox::Yes)==QMessageBox::Yes) {
      SaveChanges();
    }
  }
}
