#ifndef RFREMOTE_H
#define RFREMOTE_H

#include "ui_rfremote.h"    // Qt UIC makes this file from rfremote.ui
#include <QLabel>
//#include <QtXml/QtXml>
#include <QSettings>
#ifndef _WIN32
  #include <X11/Xlib.h>
  #include <X11/Xutil.h>
#endif
#include "qextserialport.h" // portable comm port class
#include "learn.h"

class rfremote : public QMainWindow, Ui::rfremote {
// QMainWindow is the base class of the form designed in Qt Designer (Qwidget, QDialog, QMinWindow etc.)
// Ui::rfremote is the class name of my dialog since I gave it an objectName "rfremote"
// this makes everything in QMainWindow accessable, and everything in Ui_rfremote privately accessable

  Q_OBJECT  // macro needed for signals/slots (we ALWAYS need this)

public:
  rfremote(QWidget *parent=0); // declare required class constructor, default is no parent

protected:
  void closeEvent(QCloseEvent *event); // this lets me "trap" any window close event

public slots:   // declare any slots that signals can connect to (called on button presses, menu selections, etc.)
  void doabout(void);
  void help(void);
  void Timer(void);
  void Start(void);
  void TreeMenu(const QPoint &pos);
  void AddGlobalMacro(void);
  void AddGroup(void);
  void AddMemberMacro(void);
  void EditMacro(void);
  void DeleteMacro(void);
  void LearnMacro(void);
  void ExecuteMacro(void);
  void Settings(void);
  void NewMacros(void);
  void ExportMacros(void);
  void ImportMacros(void);
  void SaveChanges(void);

public:
Display *dpy;
Window root;
  
private:
  struct macro {
    QString name;
    QString window;
    QString learn;
    QStringList commands;
    int type;
    int loop;
    int loopcnt;
  };
  void CreateActions(void);
//  QString readxmltext(QXmlStreamReader *xmlfile);
//  void readxmlrfremote(QXmlStreamReader *xmlfile);
  void readConfig(void);
  void writeConfig(void);
//  void readxmlmacro(QXmlStreamReader *xmlfile);
//  void readxmlmacros(QXmlStreamReader *xmlfile);
//  void readMacros(QString fname);
//  void writeMacros(QString fname);
  void readMacros(QSettings *settings);
  void writeMacros(QSettings *settings);
  int AddMacro(struct macro *newmacro, int parent);
  void DelMacro(int n);
  int SelectedNumber(void);
  int SelectedType(void);
  QStringList AvailableMacros(int n);
  void Show(QString name);
  Window FindWindow(Window top, const char *name);
  void SendKey(const char *keyname, bool ctrl, bool shift, bool alt, Window w);
  void ExecuteKey(QString key, QString window);
  struct macro *FindMacro(QString name);
  void ExecuteMac(QString macname);
  void ExecuteLaunch(QString launchcmd);
  void ExecuteCmd(QString cmd, QString window);
  void ExecuteIt(struct macro *macro);
  void DisplayTree(int n);
  int MatchMacro(QString data);
  void OpenPort(void);

  QString HomePath;
  QAction *addGlobalMacro;
  QAction *addGroup;
  QAction *addMemberMacro;
  QAction *editMacro;
  QAction *deleteMacro;
  QAction *learnMacro;
  QAction *executeMacro;
  // setup data
  QString port;
  int baud;
  int showpos;
  int showfontsize;
  int showfontcolor;
  int autostart;
  int autominimize;
  // end setup data
  QList<struct macro> macrolist;
  QIcon IconLearned,IconUnlearned,IconGroup;
  bool changed;
  QextSerialPort CommPort;
  QString ActiveGroup;
  QLabel *label_popshow; // popup label to show executing macro
  int timer_popshow; // timer for displaying popup show window
  bool learning;
  learn *learndialog;
  int autostart_flag;

};

#define MACRO_GLOBAL	0
#define MACRO_GROUP		1
#define MACRO_MEMBER	2

#endif
